/*************************************************************************

	File: oled.c

	Copyright (C)  S.Suwa <www.suwa-koubou.jp>

    2023.11.11  for DTMF decorder

*/

#include  <p30f2020.h>
#include  <string.h>
#include  "timer.h"
#include  "pic_delay.h"
#include  "typedef.h"
#include  "pic_232.h"
#include  "pic_print.h"
#include  "oled.h"
//#include  "font8.h"      // font size 8x8
//#include  "font8x16.h"   // font size 8x16
#include  "oled_font8x16.h"   // rotated font for 0.91'OLED

//#include  "debug.h"

/*--------------------------------------------------------------------------
	SSD1306 I2C device definition
--------------------------------------------------------------------------*/ 

#define DEVADDR    (0x3c << 1)
#define	MW         0     // I2C master write
#define MR         1     // I2C master read

#define DATA_SEND  0x40  // Control Byte Co=0, D/C#=1
#define CMD_SEND   0x00  // Control Byte Co=0, D/C#=0


//#define FONT_SIZE   6     // 5*7  X11 font 
#define FONT_SIZE   8   // 8*8  MZ-80 font


#define SWAP(x,y) { unsigned char d = x; x = y; y = d; }
#define abs(x)    ((x)>0?(x):-(x))


/*------------------------------------------------------------------------
    global variable
------------------------------------------------------------------------*/

static int font_size;

/*------------------------------------------------------------------------
    function prototype
------------------------------------------------------------------------*/

void  oled_init(void);
void  oled_clear(void);
void  oled_locate(int col, int row);
//void  oled_putstr(char *str);
void  oled_putstr2(int col, int row, char *str);
void  oled_putstr4(int col, int row, char *str);
//void  oled_putch(int c);
//void  rotate_font8x8(void);
//void  rotate_font8x16(void);

static void  oled_write_data(char *data, int len);
static void  oled_write_cmd(char *cmd, int len);


// I2C handling
static void  i2cOpenStart(void);
//static void  i2cReStart(void); // do not use
static void  i2cStopClose(void);
static int   i2cSend(char data);
//static char  i2cRec(void);
//static void  i2cAckSend(void);
//static void  i2cNAckSend(void);


/*=========================================================================
	oled_initialize
=========================================================================*/

void oled_init(void)
{
    font_size = FONT_SIZE;

    delay_ms(100);

    // 0xAE      Display Off
    // 0xA8 0x1F Set MULTIPLEX (32line)
    // 0xC0      COM output scan direction 0 to 63
    // 0xDA 0x02 Set COM Pins Hardware Configuration
    // 0x8D 0x14 Set charge pump/Enable charge pump
    // 0x81 0xCF Set Contrast Control
    // 0xAF      Display ON
    // other default value
    //oled_write_cmd("\xAE\xA8\x1F\x8D\x14\xDA\x02\x81\xFF\xAF", 10);
    oled_write_cmd("\xAE\xA8\x3F\xC0\xDA\x02\x8D\x14\x81\xFF\xAF", 11);
    oled_clear();

    //rotate_font8x8();
    //rotate_font8x16();
}

#if 0 // ********** don't use 
void rotate_font8x8(void)
{
    unsigned char font[8];
    int chr, col, row, mask, pos, data;


    for(chr=0; chr<96; chr++){
        memcpy(font, font8x8[chr], 8);

        for(col=0, mask =1; col <8; col++, mask <<=1){
            data=0;
            for(row=0, pos=0; row<8; row++, pos++){
                data |= (font[row] & mask ? 1:0) << pos;
            }
            font8x8[chr][7-col] = data;
        }
    }

}


void rotate_font8x16(void)
{
    unsigned char font[8];
    int chr, col, row, mask, pos, data;

    // upper
    for(chr=0; chr<96; chr++){
        memcpy(font, &(font8x16[chr][0]), 8);

        for(col=0, mask =1; col <8; col++, mask <<=1){
            data=0;
            for(row=0, pos=0; row<8; row++, pos++){
                data |= (font[row] & mask ? 1:0) << pos;
            }
            font8x16[chr][7-col] = data;
        }
    }

    // lower
    for(chr=0; chr<96; chr++){
        memcpy(font, &(font8x16[chr][8]), 8);

        for(col=0, mask =1; col <8; col++, mask <<=1){
            data=0;
            for(row=0, pos=0; row<8; row++, pos++){
                data |= (font[row] & mask ? 1:0) << pos;
            }
            font8x16[chr][8+7-col] = data;
        }
    }
}

#endif

/*=========================================================================
	oled_clear)()
=========================================================================*/

void oled_clear(void)
{
    int n;

    // 0x20 Set memory addressing mode
    // 0x00 Horizontal addressing mode
    // 0x21 Set column address
    // 0x00 column start address =0
    // 0x7f column end address   =127
    // 0x22 Set page address
    // 0x00 page start address =0
    // 0x03 page end address   =3
    oled_write_cmd("\x20\x00\x21\x00\x7f\x22\x00\x03", 8);

    // sending clear data
	i2cOpenStart();
	i2cSend(DEVADDR +MW);
	i2cSend(DATA_SEND);
    for(n=0; n<512; n++){ // 128column*4page
	    i2cSend(0x00);
    }
	i2cStopClose();

    oled_locate(0,0);

}


/*=========================================================================
	oled_locate()

    col: 0-20 (max 21column) at the font_size = 6
         0-15 (max 16column) at the font_size = 8
    row: 0-3  (max 4row)
=========================================================================*/

void oled_locate(int col, int row)
{
    char data[5];

    data[0]=0x20;  // Set memory addressing mode
    data[1]=0x02;  // Page addressing mode
    data[2]=0xB0 + row; // Set page start address for Page addressing mode
    data[3]= (col*font_size) & 0x0f;  // Set lower column start address for Page addressing mode
    data[4]= 0x10 + ((col*font_size) >> 4);  // Set higher column start address for Page addressing mode

    oled_write_cmd(data, 5);
}


/*=========================================================================
	oled_putstr()
    can't automatic write to next line at reached End Column.
=========================================================================*/

#if 0 // ********** don't use 
//
// using rotated font8x8
//
void oled_putstr(char *str)
{
    char data[8];

    while(*str){
        if(*str < ' ' ){ /* || *str > 0xff){ */
            memset(data, 0, font_size);
        }else{
            memcpy(data, font8x8[*str -0x20], font_size);
        }
        oled_write_data(data, font_size);
        str++;
    }
}

#endif

/*=========================================================================
	oled_putstr2()
=========================================================================*/
//
// using rotated font8x16
//
void oled_putstr2(int col, int row, char *str)
{
    char data[8];
    char *p;

    p=str;
    oled_locate(col, row);
    while(*str){
        if(*str < ' ' ){ /* || (int)*str > 0xff){ */
            memset(data, 0, font_size);
        }else{
            memcpy(data, &(font8x16[*str -0x20][0]), font_size);
        }
        oled_write_data(data, font_size);
        str++;
    }

    str = p;
    oled_locate(col, row+1);
    while(*str){
        if(*str < ' '){ /* || *str > 0xff){ */
            memset(data, 0, font_size);
        }else{
            memcpy(data, &(font8x16[*str -0x20][8]), font_size);
        }
        oled_write_data(data, font_size);
        str++;
    }
}



/*=========================================================================
	oled_putch()
    using rotated font8x8
    can't automatic write to next line at reached End Column.
=========================================================================*/
#if 0 // ********** don't use 
void oled_putch(int c)
{
    char data[8];

    if(c < ' ' || c > 0xff){
        memset(data, 0, 8);
    }else{
        memcpy(data, font8x8[c -0x20], 8);
    }
    oled_write_data(data, 6);
}
#endif

/*--------------------------------------------------------------------------
	oled_write_data()
--------------------------------------------------------------------------*/

static void oled_write_data(char *data, int len)
{
    i2cOpenStart();
    i2cSend(DEVADDR +MW);
    i2cSend(DATA_SEND);
    while(len--){
        i2cSend(*data++);
    }
    i2cStopClose();
}	


/*--------------------------------------------------------------------------
	oled_write_cmd()
--------------------------------------------------------------------------*/

static void oled_write_cmd(char *cmd, int len)
{
    i2cOpenStart();
    i2cSend(DEVADDR +MW);
    i2cSend(CMD_SEND);
    while(len--){
        i2cSend(*cmd++);
    }
    i2cStopClose();
}	


/*==========================================================================
	I2C Handling module
    dsPIC30F2020 has one i2c module.
==========================================================================*/
#define	TIME_OUT	0xffff
static	unsigned short tw;

#define I2CnCONbits  I2CCONbits
#define I2CnSTATbits I2CSTATbits
#define I2CnTRN      I2CTRN
#define I2CnRCV      I2CRCV

/*--------------------------------------------------------------------------
	I2C master Open and Start
--------------------------------------------------------------------------*/

static void	i2cOpenStart(void)
{
    // enabled I2C
    I2CnCONbits.I2CEN = 1;
    delay_us(5);         // must be need delay 

    // START 
    I2CnCONbits.SEN = 1;
    for(tw = TIME_OUT; tw > 0; tw--){
        if(!I2CnCONbits.SEN) break;
    }
}


/*--------------------------------------------------------------------------
	I2C master ReStart
--------------------------------------------------------------------------*/
#if 0  // no use this function in DTMF decorder
static void	i2cReStart(void)
{
    // RESTART 
    I2CnCONbits.RSEN = 1;
    for(tw = TIME_OUT; tw > 0; tw--){
        if(!I2CnCONbits.RSEN) break;
    }
}
#endif

/*--------------------------------------------------------------------------
	I2C master Stop and Close
--------------------------------------------------------------------------*/

static void	i2cStopClose(void)
{
    // STOP 
    I2CnCONbits.PEN = 1;
    for(tw = TIME_OUT; tw > 0; tw--){
        if(!I2CnCONbits.PEN)break;
    }

    delay_us(5);         // must be need delay
    // disabled I2C
    I2CnCONbits.I2CEN = 0;
}


/*--------------------------------------------------------------------------
	I2C master data send
--------------------------------------------------------------------------*/

static int	i2cSend(char data)
{
    I2CnTRN = data;

    if(I2CnSTATbits.IWCOL){	// If write collision occurs,return -2
        return -2;
    }

    for(tw = TIME_OUT; tw > 0; tw--){
        if(!I2CnSTATbits.TRSTAT) break;  // wait until write cycle is complete 
    }

    if (I2CnSTATbits.ACKSTAT){ 	// test for ACK condition received
        return -1;	// NACK received
    }

    return 0;	// complete with no error.
}


/*--------------------------------------------------------------------------
	I2C	master data receive
--------------------------------------------------------------------------*/
#if 0  // no use this function in DTMF decorder
static char	i2cRec(void)
{
    // enabled Receive
    I2CnCONbits.RCEN = 1;

    // wait until receive complete and data stored to I2CRCV register
    for(tw = TIME_OUT; tw > 0; tw--){
        if(!I2CnCONbits.RCEN && I2CnSTATbits.RBF) break;
    }

    // clear overflow bit
    I2CnSTATbits.I2COV = 0;
    
    return (char)I2CnRCV;
}
#endif

/*--------------------------------------------------------------------------
	I2C master Ack send
--------------------------------------------------------------------------*/
#if 0  // no use this function in DTMF decorder
static void	i2cAckSend(void)
{
    // Ack Send
    I2CnCONbits.ACKDT = 0;
    I2CnCONbits.ACKEN = 1;

    // wait until ack cycle complete
    for(tw = TIME_OUT; tw > 0; tw--){
       if(!I2CnCONbits.ACKEN)break;
    }
}
#endif

/*--------------------------------------------------------------------------
	I2C master NotAck send
--------------------------------------------------------------------------*/
#if 0  // no use this function in DTMF decorder
static void	i2cNAckSend(void)
{
    // NAck Send
    I2CnCONbits.ACKDT = 1;
    I2CnCONbits.ACKEN = 1;

    // wait until ack cycle complete
    for(tw = TIME_OUT; tw > 0; tw--){
        if(!I2CnCONbits.ACKEN) break;
    }
}
#endif

/*** end of oled.c ******************************************************/
